//
// Simple rotating 4D cube screen saver.
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>
//

#include	<windows.h>
#include	<scrnsave.h>
#include	"Vector2D.h"
#include	"Vector4D.h"

static	UINT	timerId = 0;
static	float	angle1  = 0;
static	float	angle2  = 0;
static	float	step1   = 0.05f;
static	float	step2   = 0.03f;

static	Vector4D e1 ( 1, 2, -5, 9 );
static	Vector4D e2 ( -9, 4, 2, 7 );
static	Vector4D dir ( 1, -1, 3, -1 );


HBITMAP	hBitmap = NULL;

void	orthonormalize ()
{
	float	dot    = e1 & e2;
	float	len1Sq = e1 & e1;

	e2 -= e1 * dot / len1Sq;

	e1.normalize ();
	e2.normalize ();
}

void	drawCube ( RECT& bounds, HDC hdc, COLORREF color, float a1, float a2 )
{
	HGDIOBJ	pen    = :: CreatePen ( PS_SOLID, 1, color );
	HGDIOBJ	oldPen = :: SelectObject ( hdc, pen );

	Vector2D	center ( (float)(bounds.left + bounds.right) / 2, (float)(bounds.top + bounds.bottom) / 2 );
	Vector2D	scale  ( (float)(bounds.right - bounds.left) / 7, (float)(bounds.bottom - bounds.top) / 7 );

	float	matrix [4][4];

	for ( int i = 0; i < 4; i++ )
		for ( int j = 0; j < 4; j++ )
			matrix [i][j] = 0;

	matrix [0][0] =  (float)cos ( a1 );
	matrix [1][1] =  (float)cos ( a1 );
	matrix [0][1] =  (float)sin ( a1 );
	matrix [1][0] = -(float)sin ( a1 );
	matrix [2][2] =  (float)cos ( a2 );
	matrix [3][3] =  (float)cos ( a2 );
	matrix [2][3] =  (float)sin ( a2 );
	matrix [3][2] = -(float)sin ( a2 );

	Vector4D	b1, b2;

	for ( i = 0; i < 4; i++ )
	{
		b1 [i] = 0;
		b2 [i] = 0;

		for ( int j = 0; j < 4; j++ )
		{
			b1 [i] += e1 [j] * matrix [i][j];
			b2 [i] += e2 [j] * matrix [i][j];
		}
	}

	for ( i = 0; i < 15; i++ )
	{
		Vector4D	v;

							// fill v with binary representation of i
		v.w = (i & 1 ? 1.0f : 0.0f);
		v.z = (i & 2 ? 1.0f : 0.0f);
		v.y = (i & 4 ? 1.0f : 0.0f);
		v.x = (i & 8 ? 1.0f : 0.0f);

		for ( int j = 0; j < 4; j++ )
		{
			if ( v [j] == 1 )
				continue;

			Vector4D	v2 = v;

			v2 [j] = 1;

			Vector4D	offs ( sin ( angle1 * 0.33f + angle2 * 0.2f ) * dir / 4 );
			Vector4D	w  = v  + offs;
			Vector4D	w2 = v2 + offs;
			Vector2D	start ( w  & b1,  w & b2 );
			Vector2D	end   ( w2 & b1, w2 & b2 );

							// remap projecttion to bounds
			start = center + start * scale;
			end   = center + end   * scale;

			MoveToEx ( hdc, (int)start.x, (int)start.y, NULL );
			LineTo   ( hdc, (int) end.x, (int) end.y );
		}	
	}

	SelectObject ( hdc, oldPen );
	DeleteObject ( pen );
}

void	draw ( HWND hWnd )
{
	HDC		hdc    = GetDC ( hWnd );
	HDC		hdcBuf = CreateCompatibleDC ( hdc );
	RECT	r;

    GetClientRect ( hWnd, &r );

	if ( hBitmap == NULL )
	{
					// create bitmap for double buffering
		hBitmap = CreateBitmap ( r.right - r.left, r.bottom - r.top, 1, GetDeviceCaps ( hdc, BITSPIXEL ), NULL );
	}

	HGDIOBJ	save = SelectObject ( hdcBuf, hBitmap );

    FillRect      ( hdcBuf, &r, (HBRUSH)GetStockObject ( BLACK_BRUSH ) );

	drawCube ( r, hdcBuf, RGB ( 64,  64,  64  ), angle1 - 4 * step1 / 2, angle2 - 4 * step2 / 2 );
	drawCube ( r, hdcBuf, RGB ( 64,  64,  64  ), angle1 - 3 * step1 / 2, angle2 - 3 * step2 / 2 );
	drawCube ( r, hdcBuf, RGB ( 128, 128, 128 ), angle1 - 2 * step1 / 2, angle2 - 2 * step2 / 2 );
	drawCube ( r, hdcBuf, RGB ( 192, 192, 192 ), angle1 -     step1 / 2, angle2 -     step2 );
	drawCube ( r, hdcBuf, RGB ( 255, 255, 255 ), angle1,                 angle2 );

	BitBlt ( hdc, r.left, r.top, r.right - r.left, r.bottom - r.top, hdcBuf, 0, 0, SRCCOPY );

	SelectObject ( hdcBuf, save );
	DeleteDC     ( hdcBuf );
	ReleaseDC    ( hWnd, hdc );
}

LRESULT WINAPI ScreenSaverProc ( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HDC		hDc;
	RECT	r;

	switch ( message )
    {
		case WM_CREATE:
			GetClientRect  ( hWnd, &r );
			orthonormalize ();

					// setup timer for it
			timerId = SetTimer ( hWnd, 1, 5, NULL );

            break;

		case WM_TIMER:
			angle1 += step1;
			angle2 += step2;

			draw ( hWnd );

            break;


		case WM_DESTROY:
			if ( hBitmap != NULL )
				DeleteObject ( hBitmap );

            if( timerId != 0)
				KillTimer ( hWnd, timerId );

                break;
        default:
			return DefScreenSaverProc ( hWnd, message, wParam, lParam );

	}

    return 0;
}

BOOL WINAPI ScreenSaverConfigureDialog ( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam )
{
    switch (message )
    {
        case WM_COMMAND:
			switch ( LOWORD ( wParam ) )
            {
                case IDOK:  // Exit the dialog
					EndDialog ( hDlg, 42 );
					break;
            }
			break;

        default:
		    return FALSE;
			break;
    }

	return TRUE;
}

BOOL WINAPI RegisterDialogClasses ( HANDLE hInst )
{
	return TRUE;
}

